import re
from xml.etree import ElementTree

from Empiric.internal.Print import COLORS, Print
from Empiric.internal.Statistics import VISUALIZATION_TYPE
from Empiric.Pages.Page import Page

class PageQuestionnaire(Page):
  @staticmethod
  def _attrib(n, key, defaultValue=None, mapFn=None):
    v = n.attrib[key] if n.attrib and key in n.attrib else defaultValue
    return mapFn(v) if mapFn is not None and v is not None else v
  def run(self, **settings):
    statistics = None
    if 'statistics' in settings and isinstance(settings['statistics'], str):
      statistics = {
        'title': settings['statistics'],
      }
      try:
        substatistics = {}
        if 'questions' in settings:
          questions = ElementTree.fromstring('<root>' + re.sub('required(?!=)', 'required="true"', settings['questions']) + '</root>')
          for n in questions:
            key = PageQuestionnaire._attrib(n, 'key')
            if not key or key in statistics:
              continue
            s = {'key': key}
            if n.tag == 'choice':
              substatistics[key] = {
                **s,
                'title': PageQuestionnaire._attrib(n, 'text'),
                'data': {
                  'selector': f'questionnaire.{key}',
                  'aggregateByPage': 'first',
                  'defaultValue': None,
                },
                'visualization': {
                  'type': VISUALIZATION_TYPE.BAR_CHART,
                  'options': list(map(lambda x: x.text, n)),
                },
              }
            elif n.tag == 'slider':
              substatistics[key] = {
                **s,
                'title': PageQuestionnaire._attrib(n, 'text'),
                'data': {
                  'selector': f'questionnaire.{key}',
                  'aggregateByPage': 'first',
                  'defaultValue': None,
                },
                'visualization': {
                  'type': VISUALIZATION_TYPE.BOX_PLOT,
                  'options': list(map(lambda x: x.text, n)),
                  'min': PageQuestionnaire._attrib(n, 'min', defaultValue=-2, mapFn=float),
                  'center': PageQuestionnaire._attrib(n, 'center', defaultValue=None, mapFn=float),
                  'max': PageQuestionnaire._attrib(n, 'max', defaultValue=2, mapFn=float),
                  'min-label': PageQuestionnaire._attrib(n, 'min-label'),
                  'center-label': PageQuestionnaire._attrib(n, 'center-label'),
                  'max-label': PageQuestionnaire._attrib(n, 'max-label'),
                },
              }
            elif n.tag == 'text':
              substatistics[key] = {
                **s,
                'title': PageQuestionnaire._attrib(n, 'text'),
                'data': {
                  'selector': f'questionnaire.{key}',
                  'aggregateByPage': 'first',
                },
                'visualization': {
                  'type': VISUALIZATION_TYPE.TEXT_COLLECTION,
                },
              }
        statistics['substatistics'] = list(substatistics.values())
      except Exception as e:
        Print.log('WARNING: Could not parse questionnaire: ', str(e))
    return super().run(defaultStatistics=statistics, **settings)

def pageQuestionnaire(m, title='', questions='', **kwargs):
  return PageQuestionnaire(m, 'pageQuestionnaire.html').run(title=title, questions=questions, **kwargs)



import math

from Empiric.internal.Statistics import VISUALIZATION_TYPE
from Empiric.Pages.Page import Page

class TASK:
  DRAW_GEOMETRIES = 'DRAW_GEOMETRIES'
  TRANSFORM_GEOMETRIES = 'TRANSFORM_GEOMETRIES'

class PageMap(Page):
  def run(self, **settings):
    settings = {
      'task': TASK.DRAW_GEOMETRIES,
      'transformTranslate': False,
      'transformResize': True,
      'transformResizeNonUniform': False,
      'transformRotate': True,
      'geometries': [],
      'geometriesColor': 'RED',
      'drawColor': 'BLUE',
      'drawCount': None,
      'backgroundImageSize': {'width': 2560, 'height': 1600},
      'backgroundOpacity': .6,
      'waitAfterLastDraw': 2000,
      'waitBeforeNext': 1000,
      **settings,
    }
    if 'instruction' not in settings:
      if settings['task'] == TASK.DRAW_GEOMETRIES:
        if settings['drawCount'] == 1:
          settings['instruction'] = 'Draw one additional geometry'
        else:
          settings['instruction'] = 'Draw additional geometries'
      if settings['task'] == TASK.TRANSFORM_GEOMETRIES:
        settings['instruction'] = 'Transform the geometries'
    statistics = None
    if 'statistics' in settings and isinstance(settings['statistics'], str):
      statistics = {
        'title': settings['statistics'],
        'substatistics': [],
      }
      if settings['task'] == TASK.DRAW_GEOMETRIES:
        if settings['drawCount'] is None or settings['drawCount'] > 1:
          statistics['substatistics'].append({
            'key': 'newCountdrawnGeometriesCount',
            'title': 'Number of drawn geometries',
            'data': {
              'selector': 'geometries[*].type',
              'aggregateByPage': 'count',
              'defaultValue': 0,
            },
            'visualization': {
              'type': VISUALIZATION_TYPE.BOX_PLOT,
            },
          })
        statistics['substatistics'].append({
          'key': 'drawnGeometries',
          'title': 'Drawn geometries',
          'data': {
            'selector': 'geometries[*]',
            'aggregateByPage': 'collect',
          },
          'visualization': {
            'type': VISUALIZATION_TYPE.MAP,
            'backgroundImage': settings['backgroundImage'],
            'backgroundImageSize': settings['backgroundImageSize'],
          },
        })
      if settings['task'] == TASK.TRANSFORM_GEOMETRIES:
        statistics['substatistics'].append({
          'key': 'changed',
          'title': 'Number of changed geometries',
          'data': {
            'selector': {'translate': 'geometries[*].userTranslate', 'scale': 'geometries[*].userScale', 'rotate': 'geometries[*].userRotate'},
            'aggregateByPage': 'count',
            'defaultValue': {'translate': 0, 'scale': 0, 'rotate': 0},
          },
          'visualization': {
            'type': VISUALIZATION_TYPE.BOX_PLOT,
          },
        })
        if settings['transformTranslate']:
          statistics['substatistics'].append({
            'key': 'changedTranslate',
            'title': 'Translation',
            'data': {
              'selector': 'geometries[*]',
              'aggregateByKey': 'filename',
              'value': {'x': 'userTranslate[0]', 'y': 'userTranslate[1]'},
              'defaultValue': {'x': 0, 'y': 0},
            },
            'visualization': {
              'type': VISUALIZATION_TYPE.BOX_PLOT,
              'min': -5,
              'max': 5,
            },
          })
        if settings['transformResize']:
          statistics['substatistics'].append({
            'key': 'changedScale',
            'title': 'Resize',
            'data': {
              'selector': 'geometries[*]',
              'aggregateByKey': 'filename',
              'value': {'x': 'userScale[0]', 'y': 'userScale[1]'},
              'defaultValue': {'x': 1, 'y': 1},
            },
            'visualization': {
              'type': VISUALIZATION_TYPE.BOX_PLOT,
              'min': .7,
              'max': 1.3,
            },
          })
        if settings['transformRotate']:
          statistics['substatistics'].append({
            'key': 'changedRotate',
            'title': 'Rotate',
            'data': {
              'selector': 'geometries[*]',
              'aggregateByKey': 'filename',
              'value': 'userRotate',
              'defaultValue': 0,
            },
            'visualization': {
              'type': VISUALIZATION_TYPE.BOX_PLOT,
              'min': -math.pi / 10,
              'max': math.pi / 10,
            },
          })
    return super().run(defaultStatistics=statistics, **settings)
def pageMap(m, backgroundImage=None, **kwargs):
  if not backgroundImage:
    raise Exception('Please provide a background image')
  return PageMap(m, 'pageMap.html').run(backgroundImage=backgroundImage, **kwargs)



class COLORS:
  DEFAULT = '\033[0m'
  ERROR = '\033[91m'

class Print:
  @staticmethod
  def log(*msgs):
    msg = ''.join(msgs)
    print(f' * {msg}')
  @staticmethod
  def log2(*msgs):
    msg = ''.join(msgs)
    print(f'   {msg}')
  @staticmethod
  def logWarning(*msgs):
    Print.log(COLORS.ERROR, *msgs, COLORS.DEFAULT)
  @staticmethod
  def log2Warning(*msgs):
    Print.log2(COLORS.ERROR, *msgs, COLORS.DEFAULT)



from Empiric.internal.Step import Step

class Register(Step):
  pass

def register(m, valueFn, **kwargs):
  v = valueFn()
  r = Register(m)
  r.run(raiseError=False, value=v, **kwargs)
  result = r.save({'value': v})
  return result['value']



from Empiric.internal.Print import Print

class StatisticsTools:
  @staticmethod
  def _camelCase(s):
    s = ''.join([c for c in s.replace('-', ' ') if c.isalnum() or c == ' '])
    return ''.join(x.capitalize() for x in s.split(' '))
  @staticmethod
  def fixKeyTitle(statistics):
    if 'key' not in statistics and 'title' in statistics:
      statistics['key'] = StatisticsTools._camelCase(statistics['title'])
    if 'title' not in statistics and 'key' in statistics:
      statistics['title'] = statistics['key']
    return statistics
  @staticmethod
  def statisticsToList(statistics, prefixToKey=None):
    if 'substatistics' not in statistics:
      return [statistics]
    ss = statistics['substatistics']
    if prefixToKey:
      for s in ss:
        s['key'] = prefixToKey + '-' + s['key']
    return ss
  @staticmethod
  def mergeWithDefaultStatistics(statistics, defaultStatistics, appendToDefaultStatistics, hideStatisticsByDefault):
    if not hideStatisticsByDefault and not statistics:
      Print.logWarning('Please provide \'statistics\' keyword for all relevant pages')
    if statistics is None or not defaultStatistics:
      return statistics
    if isinstance(statistics, str):
      defaultStatistics['title'] = statistics
      return defaultStatistics
    if not appendToDefaultStatistics:
      return statistics
    s = {
      'substatistics': [],
    }
    if 'key' in statistics:
      s['key'] = statistics['key']
    if 'title' in statistics:
      s['title'] = statistics['title']
    if not 'substatistics' in statistics and 'substatistics' in defaultStatistics:
      if 'key' in defaultStatistics:
        s['key'] = defaultStatistics['key']
      if 'title' in defaultStatistics:
        s['title'] = defaultStatistics['title']
    s['substatistics'].extend(StatisticsTools.statisticsToList(statistics))
    s['substatistics'].extend(StatisticsTools.statisticsToList(defaultStatistics, prefixToKey='default'))
    return s



from Empiric import Page

########################################################################
##
##  This template can be used to create new pages, which can be used in
##  combination with Empiric.
##
##  To use this template, perform the following steps:
##
##  (1) Rename the files to "pages/Page???.py" and
##      "templates/page???.html", where "???" is some string that
##      serves as a key to recognize the page.
##  (2) Rename all occurances of "pageExample" and "PageExample" in the
##      this file accordingly, while paying attention to capitalization.
##  (3) In the file where your manuscript is defined, add the following
##      import:
##      > from pages.Page??? import page???
##  (4) Now, you can use your new page in the manuscript by adding
##      > page???(m)
##  (5) Congrats, you made it!  Your first page works.
##  (6) You can start tailoring the page to your needs.  To do so,
##      modify the Python and the HTML file accordingly.
##  (7) If your page needs some JavaScript library that has not yet been
##      loaded automatically, you have to add this to the package.json,
##      which you find in your working directy.  Therafter, install the
##      JavaScript library using Yarn in your working directy:
##      > yarn install
##      Then, you can load the library accordingly by inserting
##      corresponding statements in the header block of the file
##      templates/page???.html:
##      > {% block pageHeader %}
##      >   <link rel="stylesheet" href="{{url_for('static', filename='libs/path/to/file.css')}}"/>
##      >   <script src="{{url_for('static', filename='libs/path/to/file.js')}}"></script>
##      > {% endblock %}
##
########################################################################

class PageExample(Page):
  pass

def pageExample(m, title='Example Page', message='Congrats, your page is working!', **kwargs):
  return PageExample(m, 'pageExample.html').run(title=title, message=message, **kwargs)



from Empiric.Pages.Page import Page

class PageAccessCode(Page):
  pass

def pageAccessCode(m, showErrorMessage=False, **kwargs):
  return PageAccessCode(m, 'pageAccessCode.html').render(hideStatisticsByDefault=True, showErrorMessage=showErrorMessage, **kwargs)



from Empiric.Pages.Page import Page

class PageFinal(Page):
  pass

def pageFinal(m, title='Thank you!', message='You have helped us a lot!  We highly appreciate that.', **kwargs):
  return PageFinal(m, 'pageFinal.html').run(hideStatisticsByDefault=True, title=title, message=message, **kwargs)



from flask import Flask, jsonify, redirect, render_template, request
from flask_login import login_required
import os
import shutil
import sys
import subprocess
import threading
import webbrowser

from Empiric import pkgName, pkgVersion, pkgUrl
from Empiric.internal.AccessCodes import AccessCodes
from Empiric.internal.Authenticate import Authenticate
from Empiric.internal.ManuscriptMemory import ManuscriptMemories, StepNeedsToBeRun
from Empiric.internal.Print import COLORS, Print
from Empiric.internal.Statistics import Statistics
from Empiric.Mode import MODE
from Empiric.Pages.PageAccessCode import pageAccessCode
from Empiric.Pages.PageFinal import pageFinal

class Experiment:
  _pathStaticFile = 'files'
  def __init__(self):
    self._ms = ManuscriptMemories()
    self._info()
  def _computePath(self, path, pathname):
    return path if path else os.path.join(os.path.dirname(sys.argv[0]), pathname)
  def _info(self):
    width = 54
    Print.log2()
    Print.log2('=' * width)
    Print.log2('== ', f'{pkgName} v{pkgVersion}', ' ' * (width - 8 - len(pkgName) - len(pkgVersion)), ' ==')
    Print.log2('== ', pkgUrl, ' ' * (width - 6 - len(pkgUrl)), ' ==')
    Print.log2('=' * width)
    Print.log2()
  def _yarnCheck(self):
    try:
      Print.log('Checking for Yarn')
      subprocess.run(['yarn', '--version'], capture_output=True)
      Print.log2('Found')
    except:
      Print.log2()
      Print.log2(f'{COLORS.ERROR}ERROR: Yarn is needed. Please install it.')
      Print.log2()
      Print.log2('You find Yarn here:  https://yarnpkg.com')
      Print.log2()
      Print.log2('Yarn is required to download libraries needed for the')
      Print.log2(f'web interface.{COLORS.DEFAULT}')
      Print.log2()
      return False
    return True
  def _yarnCopyPackageFiles(self, pathStatic):
    try:
      if not os.path.exists(pathStatic):
        Print.log('Creating path for static data')
        os.makedirs(pathStatic)
        Print.log2('Success')
      Print.log('Copy files to the path for static data')
      for filenameSrc, filenameDst in [('package.json', 'package.json'), ('yarn.lock', 'yarn.lock'), ('yarnrc', '.yarnrc')]:
        shutil.copy(os.path.join(os.path.dirname(__file__), 'files', filenameSrc), os.path.join(pathStatic, filenameDst))
      Print.log2('Success')
    except Exception as e:
      print(e)
      Print.log2()
      Print.log2(f'{COLORS.ERROR}ERROR: Could not copy Yarn package files.{COLORS.DEFAULT}')
      Print.log2()
      return False
    return True
  def _yarnInstall(self, pathStatic):
    try:
      Print.log('Installing JavaScript libraries using yarn')
      subprocess.run(['yarn', 'install'], capture_output=True, cwd=pathStatic)
      Print.log2('Success')
    except:
      Print.log2()
      Print.log2(f'{COLORS.ERROR}ERROR: Yarn could not install the libraries needed.')
      Print.log2()
      Print.log2('Please run yarn on your own:')
      Print.log2(f'> cd static && yarn install{COLORS.DEFAULT}')
      Print.log2()
      return False
    return True
  def _createPathStaticFile(self, pathStatic):
    try:
      p = os.path.join(pathStatic, Experiment._pathStaticFile)
      if not os.path.exists(p):
        Print.log('Creating path for static files')
        os.makedirs(p)
        Print.log2('Success')
    except:
      Print.log2()
      Print.log2(f'{COLORS.ERROR}ERROR: Could not create path for static files.{COLORS.DEFAULT}')
      Print.log2()
      return False
    return True
  def createPageStructure(self):
    pathRoot = self._computePath(None, '')
    pathPages = self._computePath(None, 'pages')
    pathPagesInit = os.path.join(pathPages, '__init__.py')
    pathTemplates = self._computePath(None, 'templates')
    if not os.path.exists(pathPages):
      Print.log('Creating path for pages')
      os.makedirs(pathPages)
      Print.log2('Success')
    if not os.path.exists(pathPagesInit):
      Print.log('Creating init file for the path for pages')
      with open(pathPagesInit, 'a'):
        pass
      Print.log2('Success')
    if not os.path.exists(pathTemplates):
      Print.log('Creating path for templates')
      os.makedirs(pathTemplates)
      Print.log2('Success')
    Print.log('Copy example files')
    for p, filename in [(pathPages, '_PageExample.py'), (pathTemplates, '_pageExample.html')]:
      if not os.path.exists(os.path.join(p, filename)):
        shutil.copy(os.path.join(os.path.dirname(__file__), 'files', filename), os.path.join(p, filename))
    for filenameSrc, filenameDst in [('package2.json', 'package.json'), ('yarnrc2', '.yarnrc')]:
      if not os.path.exists(os.path.join(pathRoot, filenameDst)):
        shutil.copy(os.path.join(os.path.dirname(__file__), 'files', filenameSrc), os.path.join(pathRoot, filenameDst))
    Print.log2('Success')
  def run(self, manuscript, port=5000, debug=False, openBrowser=True, pathStatic=None, pathTemplates=None, mode=MODE.LOCAL, numberOfAccessCodes=1000, statisticsPassword=None):
    self._port = port
    self._debug = debug
    self._openBrowser = openBrowser
    self._pathStatic = os.path.abspath(self._computePath(pathStatic, 'static'))
    self._pathTemplates = os.path.abspath(self._computePath(pathTemplates, 'templates'))
    self._mode = mode
    self._statisticsPassword = statisticsPassword
    self._readyToStart = self._yarnCheck() and self._yarnCopyPackageFiles(self._pathStatic) and self._yarnInstall(self._pathStatic) and self._createPathStaticFile(self._pathStatic)
    if not self._readyToStart:
      return
    self._ac = AccessCodes(self._mode, numberOfAccessCodes)
    app = Flask(__name__, static_folder=self._pathStatic)
    app.jinja_loader.searchpath.append(self._pathTemplates)
    Authenticate.init(app, self._statisticsPassword)
    @app.route('/')
    def base():
      if self._mode == MODE.LOCAL:
        return redirect('/' + AccessCodes.defaultAccessCode())
      elif self._mode == MODE.USE_ACCESS_CODES:
        return pageAccessCode(None)
      elif self._mode == MODE.NO_ACCESS_CODES:
        return redirect('/' + AccessCodes.newAccessCode())
    @app.route('/<string:accessCode>')
    def step(accessCode):
      if not self._ac.exists(accessCode):
        return pageAccessCode(None, showErrorMessage=True)
      m = self._ms.get(accessCode)
      m.prepareRun()
      try:
        manuscript(m)
        pageFinal(m)
      except StepNeedsToBeRun as e:
        return render_template(e.page().template(), accessCode=m.accessCode(), step=e.step())
    @app.route('/save/<string:accessCode>/<int:step>', methods=['POST'])
    def save(accessCode, step):
      if not self._ac.exists(accessCode):
        return jsonify({'success': False})
      self._ms.get(accessCode).save(step, request.get_json())
      return jsonify({'success': True})
    @app.route('/settings/<string:accessCode>/<int:step>')
    def settings(accessCode, step):
      if not self._ac.exists(accessCode):
        return jsonify({'success': False})
      return jsonify(self._ms.get(accessCode).settings(step))
    @app.route('/statistics')
    @login_required
    def statistics():
      return render_template('statistics.html')
    @app.route('/data/statistics.json')
    @login_required
    def dataStatistics():
      data = Statistics().statisticsData()
      return jsonify(data)
    if self._openBrowser:
      threading.Timer(1, lambda: webbrowser.open(f'http://127.0.0.1:{self._port}')).start()
    app.run(port=self._port, debug=self._debug)



from Empiric.Pages.Page import Page

class PageInfo(Page):
  pass

def pageInfo(m, title='', message='', **kwargs):
  return PageInfo(m, 'pageInfo.html').run(hideStatisticsByDefault=True, title=title, message=message, **kwargs)



from setuptools import setup
import setuptools

libName = 'Empiric'
libVersion = '0.0.2'
libUrl = 'https://github.com/mocnik-science/empiric'

with open('./Empiric/__info__.py', 'w') as f:
  f.write('pkgName = \'%s\'\n' % libName)
  f.write('pkgVersion = \'%s\'\n' % libVersion)
  f.write('pkgUrl = \'%s\'\n' % libUrl)

setup(
  name=libName,
  packages=['Empiric', 'Empiric.internal', 'Empiric.Pages'],
  package_data={'Empiric': ['files/*', 'templates/*']},
  install_requires=[
    'flask',
    'flask_login',
    'jsonpath-ng',
  ],
  version=libVersion,
  author='Franz-Benjamin Mocnik',
  author_email='mail@mocnik-science.net',
  description='Empiric!, an easy-to-use framework to conduct empirical experiments, with a particular focus on geospatial data perception and contribution',
  license='GPL-3',
  url=libUrl,
  download_url='',
  keywords=['empirical experiment', 'map', 'framework', 'testing', 'psychology', 'cognition'],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3',
  ],
  python_requires='>=3.6',
)



from Empiric.__info__ import pkgName, pkgVersion, pkgUrl

from Empiric.Experiment import Experiment
from Empiric.Mode import MODE
from Empiric.Pages.Page import Page
from Empiric.Pages.PageFinal import pageFinal
from Empiric.Pages.PageInfo import pageInfo
from Empiric.Pages.PageMap import pageMap, TASK
from Empiric.Pages.PageQuestionnaire import pageQuestionnaire
from Empiric.Register import register
from Empiric.Statistics import AGGREGATE, VISUALIZATION_TYPE


from datetime import datetime, timezone
import json
import os

from Empiric.internal.AccessCodes import AccessCodes
from Empiric.internal.GeneralSettings import GeneralSettings

class StepNeedsToBeRun(Exception):
  def __init__(self, step, page):
    self._step = step
    self._page = page
  def step(self):
    return self._step
  def page(self):
    return self._page

class ManuscriptMemory():
  _pathCollectedData = 'collected-data/'
  def __init__(self, accessCode):
    self._stepCounter = 0
    self._data = {
      'memory': {},
      'metadata': {
        'accessCode': accessCode,
        'timestamp': self._currentTimestamp(),
      },
    }
    self._generalSettings = GeneralSettings()
  def accessCode(self):
    return self._getMetadata('accessCode')
  def isDefaultAccessCode(self):
    return self.accessCode() == AccessCodes.defaultAccessCode()
  def _currentTimestamp(self):
    return datetime.now(tz=timezone.utc).isoformat(timespec='seconds').replace('+00:00', 'Z')
  def _filepath(self):
    if not os.path.exists(ManuscriptMemory._pathCollectedData):
      os.makedirs(ManuscriptMemory._pathCollectedData)
    tmp = self._getMetadata('timestamp').replace(':', '-') if self.isDefaultAccessCode() else self.accessCode()
    return os.path.join(ManuscriptMemory._pathCollectedData, f'experiment-{tmp}.json')
  def _memoryEmpty(self):
    return self._data['memory'] == {}
  def _stepInMemory(self, step):
    return str(step) in self._data['memory']
  def _initStep(self, step):
    self._data['memory'][str(step)] = {}
  def __getMemoryWithoutKey(self, step):
    return self._data['memory'][str(step)] if self._stepInMemory(step) else None
  def _getMemory(self, step, key):
    m = self.__getMemoryWithoutKey(step)
    return m[key] if key in m else None
  def _setMemory(self, step, key, value=None, defaultValue=None):
    m = self.__getMemoryWithoutKey(step)
    if key in m:
      if value is not None:
        m[key] = value
    else:
      m[key] = value if value is not None else defaultValue
  def _setGeneralStatistics(self, s):
    if isinstance(s, dict) and 'title' in s:
      return self._generalSettings.setStatistics(s)
    return s
  def _getMetadata(self, key):
    return self._data['metadata'][key] if key in self._data['metadata'] else None
  def prepareRun(self):
    self._stepCounter = 0
    if not self.isDefaultAccessCode() and self._memoryEmpty():
      if os.path.exists(self._filepath()):
        with open(self._filepath(), 'r') as f:
          self._data = json.load(f)
  def runStep(self, page, raiseError=True):
    self._stepCounter += 1
    if not self._stepInMemory(self._stepCounter):
      self._initStep(self._stepCounter)
    self._setMemory(self._stepCounter, 'typeOfPage', defaultValue=page.__class__.__name__)
    self._setMemory(self._stepCounter, 'settings', defaultValue=page.settings())
    self._setMemory(self._stepCounter, 'statistics', defaultValue=self._setGeneralStatistics(page.statistics()))
    result = self._getMemory(self._stepCounter, 'result')
    if result:
      return result
    if raiseError:
      raise StepNeedsToBeRun(self._stepCounter, page)
  def saveCurrentStep(self, result):
    return self.save(self._stepCounter, result)
  def save(self, step, result):
    if not step or not self._stepInMemory(step) or self._getMemory(step, 'result') is not None:
      return self._getMemory(step, 'result')
    result['timestamp'] = self._currentTimestamp()
    self._setMemory(step, 'result', defaultValue=result)
    with open(self._filepath(), 'w') as f:
      json.dump(self._data, f)
    self._generalSettings.save()
    return result
  def settings(self, step):
    settings = self._getMemory(step, 'settings')
    return settings if settings is not None else {}

class ManuscriptMemories():
  def __init__(self):
    self._ms = {}
  def get(self, accessCode):
    if accessCode not in self._ms:
      self._ms[accessCode] = ManuscriptMemory(accessCode)
    return self._ms[accessCode]



from flask import render_template

from Empiric.internal.Step import Step

class Page(Step):
  def __init__(self, m, template):
    super().__init__(m)
    self._template = template
  def run(self, **kwargs):
    return super().run(raiseError=True, **kwargs)
  def render(self, **kwargs):
    self._storeSettings(**kwargs)
    return render_template(self.template(), step=-1, **self.settings())
  def template(self):
    return self._template
